/*
   array.c -- PIKT script array routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/array.c,v 1.33 2007/01/18 13:17:31 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   -------------------------------------------------------------------------

   In the current array implementation, all variables are rendered
   internally as arrays.  Scalar values, then, are arrays with just a
   single element, and an array index of 0.  Indices for actual arrays
   begin at 1.  Doing it this way allowed us to simplify several
   stretches of code, especially in the parser.

   Values are stored in val structs, one each for valnow's (current values),
   valprc's (preceding line values), and valhst's (history values from the
   previous script run).

   In each val struct, there are 256x10 vl structs, where the actual values
   are stored.  The 256 is the maximum number of script variables permitted
   under the current implementation.  (We have ways of increasing this if
   this ever becomes an issue, but there is so much overhead associated with
   Pikt variables--especially for the @foo and %foo references--that we
   cannot easily increase the limit, or even do away with it altogether.)

   The 10 is the number of hash buckets per variable.  The hash function is
   one of the simplest imaginable:  For associative arrays, it is the first
   char of the index string modulus 10.  For numerical arrays, it is the
   numerical index % 10.

   For each of the 10 hash buckets, values are stored in circular, doubly
   linked lists.  After a successful lookup (doing a strcmp() in the case
   of associative arrays, and a simple index equality test in the case of
   numerical arrays), the list head is reassigned to the just-found node.
   In the next lookup, the search will begin at the last-found node.  If
   a match is not found at the first node, the search continues sequentially
   through the list.  A lookup fails if we come around to the first-visited
   node.  If we go more than half way around the list, a directional flag
   is set telling us to reverse the sequence of lookup next time (as when
   accessing numerical indices in decreasing order).

   Pikt numerically indexed arrays are restricted to at most three
   dimensions.  Although we might extend beyond three dimensions,
   we don't do so now for performance reasons.  Three dimensions
   should more than suffice in all but the most unusual situations.
   In those situations, you might kludge together four and more
   dimensions in an associative array with something like

   $foo["$text(ndx1) $text(ndx2) $text(ndx3) $text(ndx4) ..."]

   but you would probably be better off using another language or
   rethinking your approach to the problem.

*/

#include "../lib/common.h"
#include "piktd.h"

int hash(int, signed long *, char *, int);

/* lookup a var's spc struct; return a pointer to that struct
 */
spc *
lookupspc (int vrndx, signed long *arndx, char *arkey)
{
	static spc *sp;
	spc *initsp;
	int direction;
	int nodecnt;
	int h;
	register int k;
	int found;

	direction = DIRECTION(vrndx);
	nodecnt = 0;
	h = hash(vrndx, arndx, arkey, ARYTYP(vrndx));
	if (h == NIL)
		sp = initsp = (spc *) NULL;
	else
		sp = initsp = (alarmcurr->vars[vrndx])->datspc[h];
	while (sp) {
		nodecnt++;
		if (ARYTYP(vrndx) == STRING) {
			found = FALSE;
			if (arkey && sp->key) {
				if (! strcmp(arkey, sp->key))
					found = TRUE;
			}
		} else /* NUMBER */ {
			found = TRUE;
			for (k=0; k<=ARYDIM(vrndx); k++) {
				if (*(arndx+k) != sp->ndx[k]) {
					found = FALSE;
					break;
				}
			}
		}
		if (found)
			break;
		if (direction == FORWARD)
			sp = sp->next;
		else
			sp = sp->prev;
		if (sp == initsp) {
			sp = (spc *) NULL;
			break;
		}
	}
	if (sp) {
		if (nodecnt > (int) (((alarmcurr->vars[vrndx])->nodes[h] + 1)/2))
			DIRECTION(vrndx) = ~direction;
		(alarmcurr->vars[vrndx])->datspc[h] = sp;
	}
	return sp;
}

/* lookup a vl (value) struct; return a pointer to that struct
 */
vl *
lookupvl (val *valp, int vrndx, signed long *arndx, char *arkey, int arytyp)
{
	static vl *vlp;
	vl *initvlp;
	int direction;
	int nodecnt;
	int h;
	register int k;
	int found;

	direction = valp->direction[vrndx];
	nodecnt = 0;
	h = hash(vrndx, arndx, arkey, arytyp);
	if (h == NIL)
		vlp = initvlp = (vl *) NULL;
	else
		vlp = initvlp = valp->vals[vrndx][h];
	while (vlp) {
		nodecnt++;
		if (arytyp == STRING) {
			found = FALSE;
			if (arkey && vlp->key) {
				if (! strcmp(arkey, vlp->key))
					found = TRUE;
			}
		} else /* NUMBER */ {
			found = TRUE;
			for (k=0; k<=ARYDIM(vrndx); k++) {
				if (*(arndx+k) != vlp->ndx[k]) {
					found = FALSE;
					break;
				}
			}
		}
		if (found)
			break;
		if (direction == FORWARD)
			vlp = vlp->next;
		else
			vlp = vlp->prev;
		if (vlp == initvlp) {
			vlp = (vl *) NULL;
			break;
		}
	}
	if (vlp) {
		if (nodecnt > (int) (((valp->nodes[vrndx][h]) + 1)/2))
			valp->direction[vrndx] = ~direction;
		valp->vals[vrndx][h] = vlp;
	}
	return vlp;
}

/* find a vl struct; findvl() is a wrapper function for lookupvl(),
   allowing us in the case of valhst lookups to create the necessary
   special array lookup key, also to try a global val lookup if an
   inputline-specific lookup fails
 */
vl *
findvl (val *valp, int vrndx, signed long *arndx, char *arkey)
{
	static char fnnam[] = "findvl";
	static vl *vlp;
	char *arkeyhst;
	register int k;
	char ks[16];

	vlp = (vl *) NULL;
	if (valp == VALHST) {
		if (keystr) {
			if (arkey) {
				arkeyhst = (char *) mymalloc((keystrlen+strlen(arkey)+2)*sizeof(char), fnnam);
				sprintf(arkeyhst, "%s %s", keystr, arkey);
			} else {
				arkeyhst = (char *) mycalloc(((unsigned)keystrlen+(MAXARYDIM*16)+2), sizeof(char), fnnam);
				sprintf(arkeyhst, "%s", keystr);
				for (k=0; k<=ARYDIM(vrndx); k++) {
					sprintf(ks, " %ld", *(arndx+k));
					strcat(arkeyhst, ks);
				}
			}
			vlp = lookupvl(valp, vrndx, arndx, arkeyhst, STRING);
			free(arkeyhst);
		}
		if (! vlp) {	/* not found, so check the "globals" */
			if (arkey) {
				arkeyhst = (char *) mymalloc((keyseplen+strlen(arkey)+2)*sizeof(char), fnnam);
				sprintf(arkeyhst, "%s %s", KEYSEP, arkey);
			} else {
				arkeyhst = (char *) mycalloc(((unsigned)keyseplen+(MAXARYDIM*16)+2), sizeof(char), fnnam);
				sprintf(arkeyhst, "%s", KEYSEP);
				for (k=0; k<=ARYDIM(vrndx); k++) {
					sprintf(ks, " %ld", *(arndx+k));
					strcat(arkeyhst, ks);
				}
			}
			vlp = lookupvl(valp, vrndx, arndx, arkeyhst, STRING);
			free(arkeyhst);
		}
	} else {
		vlp = lookupvl(valp, vrndx, arndx, arkey, ARYTYP(vrndx));
	}
	return vlp;
}

/* register and initialize a new spc (spec) struct; return a pointer to that
   struct
 */
spc *
regspc (int vrndx, signed long *arndx, char *arkey)
{
	static char fnnam[] = "regspc";
	static spc *sp;
	spc *initsp;
	int h;
	register int k;

	h = hash(vrndx, arndx, arkey, ARYTYP(vrndx));
	if (h == NIL) {
		sp = (spc *) NULL;
	} else {
		sp = (spc *) mycalloc(1, sizeof(spc), fnnam);
		for (k=0; k<=MAXARYDIM; k++)
			sp->ndx[k] = *(arndx+k);
		sp->key = (((ARYTYP(vrndx) == STRING) && arkey) ? (char *) mystrdup(arkey, fnnam) : (char *)NULL);
		sp->fldnum = 0;
		sp->begincol = 0;
		sp->endcol = 0;
		sp->trim = NIL;
		initsp = (alarmcurr->vars[vrndx])->datspc[h];
		if (! initsp) {
			(alarmcurr->vars[vrndx])->datspc[h] = sp;
			sp->next = sp;
			sp->prev = sp;
		} else {
			/* (alarmcurr->vars[vrndx])->datspc[h] = sp; */
			sp->prev = initsp->prev;
			sp->next = initsp;
			(initsp->prev)->next = sp;
			initsp->prev = sp;
		}
		(alarmcurr->vars[vrndx])->nodes[h] += 1;
	}
	return sp;
}

/* register and initialize a new vl (value) struct; return a pointer to that
   struct
 */
vl *
regvl (val *valp, int vrndx, signed long *arndx, char *arkey, int arytyp)
{
	static char fnnam[] = "regvl";
	static vl *vlp;
	vl *nvlp;
	vl *initvlp;
	int h;
	char *kc;
	char *key;
	register int k;

	h = hash(vrndx, arndx, arkey, arytyp);
	if (h == NIL) {
		vlp = (vl *) NULL;
	} else {
		nvlp = (vl *) mycalloc(1, sizeof(vl), fnnam);
		for (k=0; k<=MAXARYDIM; k++)
			nvlp->ndx[k] = *(arndx+k);
		nvlp->key = (((arytyp == STRING) && arkey) ? (char *) mystrdup(arkey, fnnam) : (char *)NULL);
		nvlp->num = NIL;
		nvlp->str = (char *) NULL;
		nvlp->fil = (FILE *) NULL;
		initvlp = valp->vals[vrndx][h];
		if (! initvlp) {
			nvlp->next = nvlp;
			nvlp->prev = nvlp;
		} else {
			nvlp->prev = initvlp->prev;
			nvlp->next = initvlp;
			(initvlp->prev)->next = nvlp;
			initvlp->prev = nvlp;
		}
		valp->nodes[vrndx][h] += 1;
		vlp = valp->vals[vrndx][h] = nvlp;
		if (nvlp->key) {
			key = (char *) NULL;
			if ((valp == valnow) || (valp == valprc)) {
				key = nvlp->key;
			} else /* (valp == VALHST) */ {
				if ((kc = strchr(nvlp->key, ' ')))
					key = kc+1;
			}
			if (key) {
				/* regarkey() is such a time-consuming
				   operation (at least as currently
				   implemented; we should seek ways to
				   optimize for speed) that we avoid
				   calling it altogether unless #keys()
				   or one of the #skeys() functions applies
				   to the current var anywhere in the
				   current alarm script (as determined by
				   altlex(), alt.l)
				 */
				if (KEYSFN(vrndx))
					regarkey(vrndx, &(ARKEYS(vrndx)), arndx, key, FALSE);
				if (SORTKEYSFN(vrndx))
					regarkey(vrndx, &(SARKEYS(vrndx)), arndx, key, TRUE);
			}
		}
	}
	return vlp;
}

/* delete a vl (value) struct
 */
int
delvl (val *valp, int vrndx, signed long *arndx, char *arkey, int arytyp)
{
	static vl *vlp;
	vl *tvlp;
	vl *initvlp;
	int direction;
	int nodecnt;
	int h;
	register int k;
	int found;

	found = FALSE;
	direction = valp->direction[vrndx];
	nodecnt = 0;
	h = hash(vrndx, arndx, arkey, arytyp);
	if (h == NIL)
		vlp = initvlp = (vl *) NULL;
	else
		vlp = initvlp = valp->vals[vrndx][h];
	while (vlp) {
		nodecnt++;
		if (arytyp == STRING) {
			found = FALSE;
			if (arkey && vlp->key) {
				if (! strcmp(arkey, vlp->key))
					found = TRUE;
			}
		} else /* NUMBER */ {
			found = TRUE;
			for (k=0; k<=ARYDIM(vrndx); k++) {
				if (*(arndx+k) != vlp->ndx[k]) {
					found = FALSE;
					break;
				}
			}
		}
		if (found) {
			tvlp = vlp;
			if (vlp->next == vlp) {
				vlp = (vl *) NULL;
			} else {
				(vlp->next)->prev = vlp->prev;
				(vlp->prev)->next = vlp->next;
				vlp = vlp->next;
			}
			FREE(tvlp->str);
			free(tvlp);
			if (valp->nodes[vrndx][h])
				valp->nodes[vrndx][h] -= 1;
			break;
		}
		if (direction == FORWARD)
			vlp = vlp->next;
		else
			vlp = vlp->prev;
		if (vlp == initvlp) {
			vlp = (vl *) NULL;
			break;
		}
	}
	if (found) {
		if (nodecnt > (int) (((valp->nodes[vrndx][h]) + 1)/2))
			valp->direction[vrndx] = ~direction;
		valp->vals[vrndx][h] = vlp;
	}
	return found;
}

/* register an array key in the given keys list; if this is for #skeys(),
   insertion sort into the given list, else just add to the end; if at
   the end and sorting, assign RARKEYS() to this new key; this function
   can be very time-consuming for large data sets (very long key lists);
   we should look to optimize this for speed
 */
ky *
regarkey (int vrndx, ky **keys, signed long *arndx, char *arkey, int sort)
{
	static char fnnam[] = "regarkey";
	ky *kyp;
	ky *nkyp;
	ky *pkyp;
	int s;
	register int k;

	s = 0;
	pkyp = nkyp = (ky *) NULL;
	foreach (kyp, *keys) {
		s = strcmp(kyp->arkey, arkey);
		if (s == 0)
			return nkyp;
		else if (sort && (s < 0) && kyp->next && (strcmp(kyp->next->arkey, arkey) > 0))
			break;
		pkyp = kyp;
	}
	/* at this point, either (s < 0) or (s > 0) */
	nkyp = (ky *) mymalloc(sizeof(ky), fnnam);
	nkyp->vrndx = vrndx;
	for (k=0; k<=MAXARYDIM; k++)
		nkyp->arndx[k] = *(arndx+k);
	nkyp->arkey = mystrdup(arkey, fnnam);
	if (! *keys) {
		*keys = nkyp;
		RARKEYS(vrndx) = nkyp;
		nkyp->prev = (ky *) NULL;
		nkyp->next = (ky *) NULL;
	} else {
		if (sort) {
			if (s < 0) {
				if (kyp && kyp->next)
					(kyp->next)->prev = nkyp;
				else
					RARKEYS(vrndx) = nkyp;
				if (kyp) {
					nkyp->next = kyp->next;
					kyp->next = nkyp;
					nkyp->prev = kyp;
				} else {
					nkyp->next = (ky *) NULL;
					pkyp->next = nkyp;
					nkyp->prev = pkyp;
				}
			} else /* (s > 0) */ {
				nkyp->next = *keys;
				*keys = nkyp;
				(nkyp->next)->prev = nkyp;
				nkyp->prev = (ky *) NULL;
			}
		} else {
			pkyp->next = nkyp;
			nkyp->prev = pkyp;
			nkyp->next = (ky *) NULL;
		}
	}
	MAPKEYSCNT(vrndx) = 0;	/* signal to mapkeys() to remap the keys */
	return nkyp;
}

/* delete an array key from the given keys list
 */
int
delarkey (int vrndx, ky **keys, char *arkey, int sort)
{
	ky *kyp;
	ky *pkyp;
	int found;

	found = FALSE;
	pkyp = (ky *) NULL;
	foreach (kyp, *keys) {
		if (! strcmp(arkey, kyp->arkey)) {
			if (pkyp) {
				pkyp->next = kyp->next;
				if (kyp->next) {
					(kyp->next)->prev = pkyp;
				}
			} else {
				*keys = kyp->next;
				if (kyp->next)
					(kyp->next)->prev = (ky *) NULL;
			}
			free(kyp->arkey);
			free(kyp);
			found = TRUE;
			break;
		}
		pkyp = kyp;
	}
	if (sort && found) {
		if (pkyp && (pkyp->next == NULL))
			RARKEYS(vrndx) = pkyp;
		else if (! *keys)
			RARKEYS(vrndx) = (ky *) NULL;
	}
	MAPKEYSCNT(vrndx) = 0;	/* signal to mapkeys() to remap the keys */
	return found;
}

/* do a lookup of an array key in the given keys list
 */
ky *
lookuparkey (int vrndx, ky *keys, char *arkey)
{
	ky *kyp;

	foreach (kyp, keys) {
		if (! strcmp(kyp->arkey, arkey))
			return kyp;
	}
	return kyp;
}

/* for #members(), map the given string in the given member keys array
 */
int
mapmembers (int vrndxmem, char *mem, int first)
{
	static int memcnt;
	vl *vlp;
	register int k;

	if (first)
		memcnt = 0;
	memcnt++;
	for (k=0; k<=MAXARYDIM; k++)
		arndx[k] = 0L;
	arndx[1] = (signed long) memcnt;
	if ((vlp = lookupvl(valnow, vrndxmem, arndx, (char *)NULL, ARYTYP(vrndxmem)))) {
		FREE(vlp->str);
	} else {
		vlp = regvl(valnow, vrndxmem, arndx, (char *)NULL, ARYTYP(vrndxmem));
	}
	vlp->str = mem;
	MAPKEYSVRNDXARY(vrndxmem) = NIL;
	return memcnt;
}

/* for #skeys() and #rkeys(), map the keys for the given assoc array to
   the given member keys array
 */
int
mapkeys (int vrndxmem, ky *keys, int vrndxary, int direction)
{
	static char fnnam[] = "mapkeys";
	int keycnt;
	static ky *keyslast = (ky *) NULL;
	ky *kyp;
	vl *vlp;
	register int k;

	if ((MAPKEYSCNT(vrndxary) == 0) || (MAPKEYSVRNDXARY(vrndxmem) != vrndxary) || (keys != keyslast)) {
		delvl(valnow, vrndxmem, arndxnul, (char *)NULL, ARYTYP(vrndxmem));
		keycnt = 0;
		kyp = keys;
		while (kyp) {
			keycnt++;
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			arndx[1] = (signed long) keycnt;
			if ((vlp = lookupvl(valnow, vrndxmem, arndx, (char *)NULL, ARYTYP(vrndxmem)))) {
				FREE(vlp->str);
			} else {
				vlp = regvl(valnow, vrndxmem, arndx, (char *)NULL, ARYTYP(vrndxmem));
			}
			vlp->str = mystrdup(kyp->arkey, fnnam);
			if (direction == FORWARD)
				kyp = kyp->next;
			else /* (direction == BACKWARD) */
				kyp = kyp->prev;
		}
		MAPKEYSCNT(vrndxary) = keycnt;
		MAPKEYSVRNDXARY(vrndxmem) = vrndxary;
	} else {
		keycnt = MAPKEYSCNT(vrndxary);
	}
	keyslast = keys;
	return keycnt;
}

/* an exceedingly simple hash determination; in the case of an array key,
   the key's first char % 10; in the case of an array index, loop through
   all the indices, summing the current index with the previous sum and
   performing a % 10 against the new sum
 */
int
hash (int vrndx, signed long *arndx, char *arkey, int arytyp)
{
	register int h;
	register int k;

	if (arytyp == STRING) {
		if (arkey && *arkey) {
			h = (int) *arkey % 10;
		} else {
			h = NIL;
		}
	} else /* NUMBER */ {
		h = 0;
		for (k=0; k<=ARYDIM(vrndx); k++) {
			if (*(arndx+k) >= 0L) {
				h = (h + *(arndx+k) % 10) % 10;
			} else {
				h = NIL;
				break;
			}
		}
	}
	return h;
}
